# pylint: disable=missing-docstring

from nengo import builder
from nengo.builder import signal, operator
import numpy as np
import pytest
import tensorflow as tf

from nengo_dl.op_builders import sparse_matmul
from nengo_dl.utils import tf_gpu_installed


def test_elementwise_inc(Simulator):
    # note: normally the op_builders are just tested as part of the nengo
    # tests.  but in this particular case, there are no nengo tests that
    # have a scalar, non-1 transform.  those all get optimized out during
    # the graph optimization, so we don't end up with any tests of
    # elementwiseinc where A is a scalar. so that's what this is for.

    model = builder.Model()

    a = signal.Signal([2.0])
    x = signal.Signal([[3.0]])
    y = signal.Signal([[1.0]])
    op = operator.ElementwiseInc(a, x, y)
    model.add_op(op)

    with Simulator(None, model=model) as sim:
        sim.sess.run(
            sim.tensor_graph.steps_run,
            feed_dict={sim.tensor_graph.step_var: 0, sim.tensor_graph.stop_var: 5},
        )


@pytest.mark.parametrize("device", ("/cpu:0", "/gpu:0", None))
@pytest.mark.parametrize("dtype", (tf.float32, tf.float64))
def test_sparse_matmul(sess, dtype, device):
    if device == "/gpu:0" and not tf_gpu_installed:
        pytest.skip("Can't test GPU device without GPU support")

    with tf.device(device):
        A = tf.ones(3, dtype=dtype)
        idxs = tf.constant([[0, 0], [1, 1], [2, 2]])
        shape = (3, 3)
        X = tf.ones((3, 1), dtype=dtype) * 2

        with pytest.warns(None) as recwarns:
            Y = sparse_matmul(idxs, A, shape, X)

    if dtype == tf.float64 and (
        device == "/gpu:0" or (device is None and tf_gpu_installed)
    ):
        assert len([w for w in recwarns if "sparse_matmul" in str(w.message)]) == 1

        # need to go back past cast
        assert Y.op.inputs[0].op.inputs[1].dtype == tf.float32
    else:
        assert len([w for w in recwarns if "sparse_matmul" in str(w.message)]) == 0
        assert Y.op.inputs[1].dtype == dtype

    assert Y.dtype == dtype

    assert np.allclose(sess.run(Y), np.ones(3) * 2)



# pylint: disable=missing-docstring

from nengo.builder.neurons import SimNeurons
from nengo.builder.signal import Signal
from nengo.exceptions import BuildError
from nengo.neurons import LIF
import numpy as np
import pytest
import tensorflow as tf

from nengo_dl.compat import tf_compat, RefVariable
from nengo_dl.signals import TensorSignal, SignalDict


def test_tensor_signal_basic():
    # check that indices are read-only
    sig = TensorSignal([0, 1, 2], None, None, None, None, None)
    with pytest.raises(BuildError):
        sig.indices = [2, 3, 4]
    with pytest.raises(ValueError):
        sig.indices[0] = 1

    # check ndim
    sig = TensorSignal([0, 1, 2], None, None, (1, 2), 1, None)
    assert sig.ndim == 2


def test_tensor_signal_getitem():
    sig = TensorSignal([1, 2, 3, 4], object(), None, (4, 3), None, None)
    sig_slice = sig[:2]
    assert np.all(sig_slice.indices == (1, 2))
    assert sig_slice.key == sig.key
    assert sig_slice.shape == (2, 3)

    assert sig[...] is sig

    sig_adv = sig[[1, 3]]
    assert np.all(sig_adv.indices == (2, 4))
    assert sig_adv.shape == (2, 3)


def test_tensor_signal_reshape():
    sig = TensorSignal([1, 2, 3, 4], object(), None, (4, 3), None, None)

    with pytest.raises(BuildError):
        sig.reshape((100,))

    sig_reshape = sig.reshape((6, 2))
    assert np.all(sig_reshape.indices == sig.indices)
    assert sig_reshape.key == sig.key
    assert sig_reshape.shape == (6, 2)

    sig_reshape = sig.reshape((6, 2, 1))
    assert np.all(sig_reshape.indices == sig.indices)
    assert sig_reshape.key == sig.key
    assert sig_reshape.shape == (6, 2, 1)

    sig_reshape = sig.reshape((-1, 2))
    assert sig_reshape.shape == (6, 2)

    sig_reshape = sig.reshape((-1,))
    assert sig_reshape.shape == (12,)

    with pytest.raises(BuildError):
        sig.reshape((-1, 5))

    with pytest.raises(BuildError):
        sig.reshape((-1, -1))

    with pytest.raises(BuildError):
        sig.reshape((4, 4))


def test_tensor_signal_broadcast():
    sig = TensorSignal([0, 1, 2, 3], object(), None, (4,), None, None)
    base = np.random.randn(4)

    sig_broad = sig.broadcast(-1, 2)
    assert sig_broad.shape == (4, 2)
    assert sig_broad.key == sig.key
    assert np.all(np.reshape(base[sig_broad.indices], sig_broad.shape) == base[:, None])

    sig_broad = sig.broadcast(0, 2)
    assert sig_broad.shape == (2, 4)
    assert sig_broad.key == sig.key
    assert np.all(np.reshape(base[sig_broad.indices], sig_broad.shape) == base[None, :])


def test_tensor_signal_load_indices(sess):
    sig = TensorSignal([2, 3, 4, 5], object(), None, (4,), None, tf.constant)
    assert np.all(sess.run(sig.tf_indices) == sig.indices)
    start, stop, step = sess.run(sig.tf_slice)
    assert start == 2
    assert stop == 6
    assert step == 1

    sig = TensorSignal([2, 4, 6, 8], object(), None, (4,), None, tf.constant)
    assert np.all(sess.run(sig.tf_indices) == sig.indices)
    start, stop, step = sess.run(sig.tf_slice)
    assert start == 2
    assert stop == 9
    assert step == 2

    sig = TensorSignal([2, 2, 3, 3], object(), None, (4,), None, tf.constant)
    assert np.all(sess.run(sig.tf_indices) == sig.indices)
    assert sig.tf_slice is None


def test_signal_dict_scatter(sess):
    minibatch_size = 1
    var_size = 19
    signals = SignalDict(tf.float32, minibatch_size)

    key = object()
    val = np.random.randn(var_size, minibatch_size)
    signals.bases = {key: tf_compat.assign(RefVariable(val, dtype=tf.float32), val)}

    x = signals.get_tensor_signal([0, 1, 2, 3], key, tf.float32, (4,), True)
    with pytest.raises(BuildError):
        # wrong dtype
        signals.scatter(x, tf.ones((4,), dtype=tf.float64))

    # update
    signals.scatter(x, tf.ones((4,)))
    y = sess.run(signals.bases[key])
    assert np.allclose(y[:4], 1)
    assert np.allclose(y[4:], val[4:])

    # increment, and reshaping val
    signals.scatter(x, tf.ones((2, 2)), mode="inc")
    y = sess.run(signals.bases[key])
    assert np.allclose(y[:4], 2)
    assert np.allclose(y[4:], val[4:])

    # recognize assignment to full array
    x = signals.get_tensor_signal(
        np.arange(var_size), key, tf.float32, (var_size,), True
    )
    y = tf.ones((var_size, 1))
    signals.scatter(x, y)
    assert signals.bases[key].op.type == "Assign"

    # recognize assignment to strided full array
    x = signals.get_tensor_signal(
        np.arange(0, var_size, 2), key, tf.float32, (var_size // 2 + 1,), True
    )
    y = tf.ones((var_size // 2 + 1, 1))
    signals.scatter(x, y)
    assert signals.bases[key].op.type == "ScatterUpdate"


def test_signal_dict_gather(sess):
    minibatch_size = 1
    var_size = 19
    signals = SignalDict(tf.float32, minibatch_size)

    key = object()
    val = np.random.randn(var_size, minibatch_size)
    signals.bases = {key: tf.constant(val, dtype=tf.float32)}

    x = signals.get_tensor_signal([0, 1, 2, 3], key, tf.float32, (4,), True)

    # sliced read
    assert np.allclose(sess.run(signals.gather(x)), val[:4])

    # read with reshape
    x = signals.get_tensor_signal([0, 1, 2, 3], key, tf.float32, (2, 2), True)
    assert np.allclose(
        sess.run(signals.gather(x)), val[:4].reshape((2, 2, minibatch_size))
    )

    # gather read
    x = signals.get_tensor_signal([0, 1, 2, 3], key, tf.float32, (4,), True)
    y = signals.gather(x, force_copy=True)
    assert "Gather" in y.op.type

    x = signals.get_tensor_signal([0, 0, 3, 3], key, tf.float32, (4,), True)
    assert np.allclose(sess.run(signals.gather(x)), val[[0, 0, 3, 3]])
    assert "Gather" in y.op.type

    # reading from full array
    x = signals.get_tensor_signal(
        np.arange(var_size), key, tf.float32, (var_size,), True
    )
    y = signals.gather(x)
    assert y is signals.bases[key]

    # reading from strided full array
    x = signals.get_tensor_signal(
        np.arange(0, var_size, 2), key, tf.float32, (var_size // 2 + 1,), True
    )
    y = signals.gather(x)
    assert y.op.type == "StridedSlice"
    assert y.op.inputs[0] is signals.bases[key]

    # minibatch dimension
    x = signals.get_tensor_signal([0, 1, 2, 3], key, tf.float32, (4,), True)
    assert signals.gather(x).get_shape() == (4, 1)

    x = signals.get_tensor_signal([0, 1, 2, 3], key, tf.float32, (4,), False)
    assert signals.gather(x).get_shape() == (4,)


def test_signal_dict_combine():
    minibatch_size = 1
    signals = SignalDict(tf.float32, minibatch_size)

    key = object()

    assert signals.combine([]) == []

    y = signals.combine(
        [
            signals.get_tensor_signal([0, 1, 2], key, None, (3, 2), False),
            signals.get_tensor_signal([4, 5, 6], key, None, (3, 2), False),
        ]
    )
    assert y.key is key
    assert y._tf_indices is None

    assert y.shape == (6, 2)

    assert np.all(y.indices == [0, 1, 2, 4, 5, 6])


@pytest.mark.parametrize("dtype", (None, tf.float32))
def test_constant(dtype, sess):
    val = np.random.randn(10, 10).astype(np.float64)

    signals = SignalDict(tf.float32, 1)
    const0 = signals.constant(val, dtype=dtype)
    const1 = signals.constant(val, dtype=dtype, cutoff=0)

    assert const0.op.type == "Const"
    assert const1.op.type == "Identity"

    assert const0.dtype == (dtype if dtype else tf.as_dtype(val.dtype))
    assert const1.dtype == (dtype if dtype else tf.as_dtype(val.dtype))

    sess.run(
        tf_compat.variables_initializer(tf_compat.get_collection("constants")),
        feed_dict=signals.constant_phs,
    )
    c0, c1 = sess.run([const0, const1])

    assert np.allclose(c0, val)
    assert np.allclose(c1, val)


@pytest.mark.gpu
def test_constant_gpu(sess):
    val = np.random.randn(10, 10).astype(np.int32)

    with tf.device("/gpu:0"):
        signals = SignalDict(tf.float32, 1)
        const = signals.constant(val, cutoff=0)

        assert const.dtype == tf.int32
        assert "GPU" in const.device.upper()

    sess.run(
        tf_compat.variables_initializer(tf_compat.get_collection("constants")),
        feed_dict=signals.constant_phs,
    )
    c = sess.run(const)

    assert np.allclose(val, c)


@pytest.mark.parametrize("dtype", (np.float32, np.float64))
@pytest.mark.parametrize("diff", (True, False))
def test_op_constant(dtype, diff, sess):
    ops = (
        SimNeurons(LIF(tau_rc=1), Signal(np.zeros(10)), None),
        SimNeurons(LIF(tau_rc=2 if diff else 1), Signal(np.zeros(10)), None),
    )

    signals = SignalDict(tf.float32, 1)
    const = signals.op_constant(
        [op.neurons for op in ops], [op.J.shape[0] for op in ops], "tau_rc", dtype
    )
    const1 = signals.op_constant(
        [op.neurons for op in ops],
        [op.J.shape[0] for op in ops],
        "tau_rc",
        dtype,
        ndims=1,
    )
    const3 = signals.op_constant(
        [op.neurons for op in ops],
        [op.J.shape[0] for op in ops],
        "tau_rc",
        dtype,
        ndims=3,
    )

    assert const.dtype.base_dtype == dtype

    sess.run(
        tf_compat.variables_initializer(tf_compat.get_collection("constants")),
        feed_dict=signals.constant_phs,
    )
    x, x1, x3 = sess.run([const, const1, const3])

    if diff:
        assert np.array_equal(x, [[1]] * 10 + [[2]] * 10)
        assert np.array_equal(x[:, 0], x1)
        assert np.array_equal(x, x3[..., 0])
    else:
        assert np.array_equal(x, 1.0)
        assert np.array_equal(x, x1)
        assert np.array_equal(x, x3)


def test_get_tensor_signal():
    signals = SignalDict(tf.float32, 3)

    # check that tensor_signal is created correctly
    key = object()
    tensor_signal = signals.get_tensor_signal((0,), key, np.float64, (3, 4), True)

    assert isinstance(tensor_signal, TensorSignal)
    assert np.array_equal(tensor_signal.indices, (0,))
    assert tensor_signal.key == key
    assert tensor_signal.dtype == np.float64
    assert tensor_signal.shape == (3, 4)
    assert tensor_signal.minibatch_size == 3
    assert tensor_signal.constant == signals.constant
    assert len(signals) == 0

    # check adding signal to sig_map
    sig = Signal(np.zeros(4))
    sig.minibatched = True
    tensor_signal = signals.get_tensor_signal(
        np.arange(4), key, np.float64, (2, 2), True, signal=sig
    )
    assert len(signals) == 1
    assert signals[sig] is tensor_signal
    assert next(iter(signals)) is sig
    assert next(iter(signals.values())) is tensor_signal

    # error if sig shape doesn't match indices
    with pytest.raises(AssertionError):
        sig = Signal(np.zeros((2, 2)))
        sig.minibatched = True
        signals.get_tensor_signal(
            np.arange(4), key, np.float64, (2, 2), True, signal=sig
        )

    # error if sig size doesn't match given shape
    with pytest.raises(AssertionError):
        sig = Signal(np.zeros(4))
        sig.minibatched = True
        signals.get_tensor_signal(
            np.arange(4), key, np.float64, (2, 3), True, signal=sig
        )

    # error if minibatched doesn't match
    with pytest.raises(AssertionError):
        sig = Signal(np.zeros(4))
        sig.minibatched = False
        signals.get_tensor_signal(
            np.arange(4), key, np.float64, (2, 2), True, signal=sig
        )



from distutils.version import LooseVersion
import shlex

import nengo
from nengo.tests import test_synapses, test_learning_rules
import numpy as np
import pytest
import tensorflow as tf

from nengo_dl import simulator, utils
from nengo_dl.compat import tf_compat
from nengo_dl.tests import make_test_sim


def pytest_configure(config):
    # add unsupported attribute to Simulator (for compatibility with nengo<3.0)
    # join all the lines and then split (preserving quoted strings)
    if nengo.version.version_info <= (2, 8, 0):
        unsupported = shlex.split(" ".join(config.getini("nengo_test_unsupported")))
        # group pairs (representing testname + reason)
        unsupported = [unsupported[i : i + 2] for i in range(0, len(unsupported), 2)]
        # wrap square brackets to interpret them literally
        # (see https://docs.python.org/3/library/fnmatch.html)
        for i, (testname, _) in enumerate(unsupported):
            unsupported[i][0] = "".join(
                "[%s]" % c if c in ("[", "]") else c for c in testname
            ).replace("::", ":")

        simulator.Simulator.unsupported = unsupported


def pytest_runtest_setup(item):
    if item.get_closest_marker("gpu", False) and not utils.tf_gpu_installed:
        pytest.skip("This test requires tensorflow-gpu")
    elif (
        hasattr(item, "fixturenames")
        and "Simulator" not in item.fixturenames
        and item.config.getvalue("--simulator-only")
    ):
        pytest.skip("Only running tests that require a Simulator")
    elif item.get_closest_marker("training", False) and item.config.getvalue(
        "--inference-only"
    ):
        pytest.skip("Skipping training test in inference-only mode")
    elif item.get_closest_marker("performance", False) and not item.config.getvalue(
        "--performance"
    ):
        pytest.skip("Skipping performance test")


def pytest_addoption(parser):
    parser.addoption(
        "--simulator-only",
        action="store_true",
        default=False,
        help="Only run tests involving Simulator",
    )
    parser.addoption(
        "--inference-only",
        action="store_true",
        default=False,
        help="Run tests in inference-only mode",
    )
    parser.addoption(
        "--dtype",
        default="float32",
        choices=("float32", "float64"),
        help="Simulator float precision",
    )
    parser.addoption(
        "--unroll-simulation",
        default=1,
        type=int,
        help="`unroll_simulation` parameter for Simulator",
    )
    parser.addoption("--device", default=None, help="`device` parameter for Simulator")
    parser.addoption(
        "--performance",
        action="store_true",
        default=False,
        help="Run performance tests",
    )

    if nengo.version.version_info <= (2, 8, 0):
        # add the pytest options from future nengo versions
        parser.addini(
            "nengo_test_unsupported",
            type="linelist",
            help="List of unsupported unit tests with reason for exclusion",
        )


@pytest.fixture(scope="session")
def Simulator(request):
    """
    Simulator class to be used in tests (use this instead of ``nengo_dl.Simulator``).
    """

    return make_test_sim(request)


@pytest.fixture(scope="function")
def sess(request):
    """Create a TensorFlow session with a unique scope per test."""
    with tf.Graph().as_default() as graph:
        with tf_compat.Session(graph=graph) as sess:
            yield sess


def patch_nengo_tests():
    """
    Monkey-patch various aspects of the Nengo test suite, so that things
    work correctly when running those tests through NengoDL.
    """

    if LooseVersion(nengo.__version__) < "3.0.0":
        from nengo import conftest

        # monkey patch the nengo Simulator fixture, so that we can also use the pytest
        # arguments to control nengo tests
        conftest.Simulator = Simulator

        # set looser tolerances on synapse tests (since allclose fixture doesn't work
        # in these versions)
        def allclose(*args, **kwargs):
            kwargs.setdefault("atol", 5e-5)
            return nengo.utils.testing.allclose(*args, **kwargs)

        test_synapses.allclose = allclose

        # cast output of run_synapse to float64. this is necessary because
        # Synapse.filt bases its internal dtypes on the dtype of its inputs, and
        # we don't want to downcast everything there to float32.
        nengo_run_synapse = test_synapses.run_synapse

        def run_synapse(*args, **kwargs):
            output = nengo_run_synapse(*args, **kwargs)
            return tuple(x.astype(np.float64) for x in output)

        test_synapses.run_synapse = run_synapse

        # remove the correction probes from _test_pes
        def _test_pes(
            NengoSimulator,
            nl,
            plt,
            seed,
            pre_neurons=False,
            post_neurons=False,
            weight_solver=False,
            vin=np.array([0.5, -0.5]),
            vout=None,
            n=200,
            function=None,
            transform=np.array(1.0),
            rate=1e-3,
        ):
            vout = np.array(vin) if vout is None else vout

            model = nengo.Network(seed=seed)
            with model:
                model.config[nengo.Ensemble].neuron_type = nl()

                u = nengo.Node(output=vin)
                v = nengo.Node(output=vout)
                a = nengo.Ensemble(n, dimensions=u.size_out)
                b = nengo.Ensemble(n, dimensions=u.size_out)
                e = nengo.Ensemble(n, dimensions=v.size_out)

                nengo.Connection(u, a)

                bslice = b[: v.size_out] if v.size_out < u.size_out else b
                pre = a.neurons if pre_neurons else a
                post = b.neurons if post_neurons else bslice

                conn = nengo.Connection(
                    pre,
                    post,
                    function=function,
                    transform=transform,
                    learning_rule_type=nengo.PES(rate),
                )
                if weight_solver:
                    conn.solver = nengo.solvers.LstsqL2(weights=True)

                nengo.Connection(v, e, transform=-1)
                nengo.Connection(bslice, e)
                nengo.Connection(e, conn.learning_rule)

                b_p = nengo.Probe(bslice, synapse=0.03)
                e_p = nengo.Probe(e, synapse=0.03)

                weights_p = nengo.Probe(conn, "weights", sample_every=0.01)

            with NengoSimulator(model) as sim:
                sim.run(0.5)
            t = sim.trange()
            weights = sim.data[weights_p]

            tend = t > 0.4
            assert np.allclose(sim.data[b_p][tend], vout, atol=0.05)
            assert np.allclose(sim.data[e_p][tend], 0, atol=0.05)
            assert not np.allclose(weights[0], weights[-1], atol=1e-5)

        test_learning_rules._test_pes = _test_pes


patch_nengo_tests()



# pylint: disable=wrong-import-order,wrong-import-position,missing-docstring,ungrouped-imports

__copyright__ = "2015-2018, Applied Brain Research"
__license__ = "Free for non-commercial use; see LICENSE.rst"
from nengo_dl.version import version as __version__

# check python version
import sys

if sys.version_info < (3, 5):
    raise ImportError(
        """
You are running Python version %s with NengoDL version %s. NengoDL requires
at least Python 3.5.

The fact that this version was installed on your system probably means that you
are using an older version of pip; you should consider upgrading with

 $ pip install pip setuptools --upgrade

There are two options for getting NengoDL working:

- Upgrade to Python >= 3.5

- Install an older version of NengoDL:

 $ pip install 'nengo-dl<2.0'
"""
        % (sys.version, __version__)
    )
del sys

# filter out "INFO" level log messages
import os
from nengo_dl.compat import tf_compat

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"
tf_compat.logging.set_verbosity(tf_compat.logging.WARN)


# disable v2 behaviour (for now, pending full TF 2.0 compatibility)
try:
    tf_compat.disable_v2_behavior()
except AttributeError:
    # on a version of TensorFlow that doesn't have v2 behaviour anyway, so no problem
    pass


del os
del tf_compat

# need to explicitly import these to trigger the builder registration
from nengo_dl import (
    op_builders,
    neuron_builders,
    process_builders,
    learning_rule_builders,
    transform_builders,
)

# import into top-level namespace
from nengo_dl import dists, objectives
from nengo_dl.simulator import Simulator
from nengo_dl.tensor_node import TensorNode, tensor_layer, reshaped
from nengo_dl.config import configure_settings
from nengo_dl.neurons import SoftLIFRate

# shortcut for objective namespace
obj = objectives

# apply tensorflow monkey patches
from nengo_dl import tensorflow_patch

tensorflow_patch.patch_dynamic_stitch_grad()
tensorflow_patch.patch_state_grads()



# pylint: disable=unused-import,ungrouped-imports

"""
Utilities to ease cross-compatibility between different versions of upstream
dependencies.
"""

from distutils.version import LooseVersion

import nengo
import numpy as np
import tensorflow as tf

if LooseVersion(tf.__version__) < LooseVersion("2.0.0"):
    tf_compat = tf

    def tf_convolution(*args, **kwargs):
        """Convert "filters" kwarg to "filter"."""

        if "filters" in kwargs:
            kwargs["filter"] = kwargs.pop("filters")
        return tf.nn.convolution(*args, **kwargs)

    tf_math = tf

    tf_uniform = tf.random_uniform

    RefVariable = tf.Variable
else:
    tf_compat = tf.compat.v1

    tf_convolution = tf.nn.convolution

    tf_math = tf.math

    tf_uniform = tf.random.uniform

    def RefVariable(*args, **kwargs):
        """Always returns RefVariables instead of (maybe) ResourceVariables."""

        return tf.compat.v1.Variable(*args, use_resource=False, **kwargs)


if LooseVersion(nengo.__version__) < "3.0.0":
    from nengo.utils.testing import allclose as signals_allclose

    class SimPES(nengo.builder.Operator):  # pylint: disable=abstract-method
        """Future `nengo.builder.operator.SimPES` class."""

        def __init__(
            self, pre_filtered, error, delta, learning_rate, encoders=None, tag=None
        ):
            super(SimPES, self).__init__(tag=tag)

            self.pre_filtered = pre_filtered
            self.error = error
            self.delta = delta
            self.learning_rate = learning_rate
            self.encoders = encoders

            # encoders not used in NengoDL (they'll be applied outside the op)
            assert encoders is None

            # note: in 3.0.0 the PES op changed from a set to an update
            self.sets = [delta]
            self.incs = []
            self.reads = [pre_filtered, error]
            self.updates = []

        def _descstr(self):
            return "pre=%s, error=%s -> %s" % (
                self.pre_filtered,
                self.error,
                self.delta,
            )

    # remove 'correction' from probeable attributes
    nengo.PES.probeable = ("error", "activities", "delta")

    class Convolution:
        """Dummy `nengo.transforms.Convolution` class."""

    class ConvInc:
        """Dummy `nengo.builder.transforms.ConvInc` class."""

    class SparseMatrix:
        """Dummy `nengo.transforms.SparseMatrix` class."""

    class SparseDotInc:
        """Dummy `nengo.builder.transforms.SparseDotInc` class."""

    def is_sparse(sig):
        """Check if Signal is sparse"""
        # always False since Sparse signals didn't exist, but we use getattr
        # so that dummies.Signal(sparse=False) will still work
        return getattr(sig, "sparse", False)

    class Step:
        """Future `nengo.synapses.LinearFilter.Step` class."""

        def __init__(self, A, B, C, D):
            self.A = A
            self.B = B
            self.C = C
            self.D = D

        @staticmethod
        def check(A, B, C, D):
            """Check if the given matrices represent a system of this type."""

    class NoX(Step):
        """Future `nengo.synapses.LinearFilter.NoX` class."""

        @staticmethod
        def check(A, B, C, D):
            return A.size == 0

    class OneX(Step):
        """Future `nengo.synapses.LinearFilter.OneX` class."""

        @staticmethod
        def check(A, B, C, D):
            return len(A) == 1 and (D == 0).all()

    class NoD(Step):
        """Future `nengo.synapses.LinearFilter.NoD` class."""

        @staticmethod
        def check(A, B, C, D):
            return len(A) >= 1 and (D == 0).all()

    class General(Step):
        """Future `nengo.synapses.LinearFilter.General` class."""

        @staticmethod
        def check(A, B, C, D):
            return len(A) >= 1

    def make_linear_step(synapse, input_shape, output_shape, dt):
        """Generate one of the future LinearFilter step types."""
        # pylint: disable=import-outside-toplevel
        from nengo.utils.filter_design import cont2discrete, tf2ss

        A, B, C, D = tf2ss(synapse.num, synapse.den)

        # discretize (if len(A) == 0, filter is stateless and already discrete)
        if synapse.analog and len(A) > 0:
            A, B, C, D, _ = cont2discrete((A, B, C, D), dt, method="zoh")

        if NoX.check(A, B, C, D):
            return NoX(A, B, C, D)
        elif OneX.check(A, B, C, D):
            return OneX(A, B, C, D)
        elif NoD.check(A, B, C, D):
            return NoD(A, B, C, D)
        else:
            assert General.check(A, B, C, D)
            return General(A, B, C, D)

    # monkey-patch in empty state attribute
    from nengo.builder.processes import SimProcess

    SimProcess.state = {}

    def make_process_step(process, shape_in, shape_out, dt, rng, _):
        """Ignore state argument."""
        return process.make_step(shape_in, shape_out, dt, rng)

    def make_process_state(process, shape_in, shape_out, dt):
        """Old processes don't have any state."""
        return {}

    # set seed upper bound to uint32 (instead of int32)
    nengo.base.Process.seed.high = np.iinfo(np.uint32).max

else:
    from nengo.builder.learning_rules import SimPES
    from nengo.builder.transforms import ConvInc, SparseDotInc
    from nengo.transforms import Convolution, SparseMatrix
    from nengo.synapses import LinearFilter
    from nengo.utils.testing import signals_allclose

    def is_sparse(sig):
        """Check if Signal is sparse"""
        return sig.sparse

    NoX = LinearFilter.NoX
    OneX = LinearFilter.OneX
    NoD = LinearFilter.NoD
    General = LinearFilter.General

    def make_linear_step(synapse, input_shape, output_shape, dt):
        """Call synapse.make_step to compute A/B/C/D."""
        state = synapse.make_state(input_shape, output_shape, dt)
        return synapse.make_step(input_shape, output_shape, dt, rng=None, state=state)

    def make_process_step(process, shape_in, shape_out, dt, rng, state):
        """Call process.make_step."""
        return process.make_step(shape_in, shape_out, dt, rng, state)

    def make_process_state(process, shape_in, shape_out, dt):
        """Call process.make_state."""
        return process.make_state(shape_in, shape_out, dt)



"""
Benchmark networks and utilities for evaluating NengoDL's performance.
"""

import inspect
import itertools
import os
import random
import time

import click
import matplotlib.pyplot as plt
import nengo
import numpy as np
import tensorflow as tf

import nengo_dl
from nengo_dl.compat import tf_compat


def cconv(dimensions, neurons_per_d, neuron_type):
    """
    Circular convolution (EnsembleArray) benchmark.

    Parameters
    ----------
    dimensions : int
        Number of dimensions for vector values
    neurons_per_d : int
        Number of neurons to use per vector dimension
    neuron_type : `~nengo.neurons.NeuronType`
        Simulation neuron type

    Returns
    -------
    net : `nengo.Network`
        benchmark network
    """

    with nengo.Network(label="cconv", seed=0) as net:
        net.config[nengo.Ensemble].neuron_type = neuron_type
        net.config[nengo.Ensemble].gain = nengo.dists.Choice([1, -1])
        net.config[nengo.Ensemble].bias = nengo.dists.Uniform(-1, 1)

        net.cconv = nengo.networks.CircularConvolution(neurons_per_d, dimensions)

        net.inp_a = nengo.Node([0] * dimensions)
        net.inp_b = nengo.Node([1] * dimensions)
        nengo.Connection(net.inp_a, net.cconv.A)
        nengo.Connection(net.inp_b, net.cconv.B)

        net.p = nengo.Probe(net.cconv.output)

    return net


def integrator(dimensions, neurons_per_d, neuron_type):
    """
    Single integrator ensemble benchmark.

    Parameters
    ----------
    dimensions : int
        Number of dimensions for vector values
    neurons_per_d : int
        Number of neurons to use per vector dimension
    neuron_type : `~nengo.neurons.NeuronType`
        Simulation neuron type

    Returns
    -------
    net : `nengo.Network`
        benchmark network
    """

    with nengo.Network(label="integrator", seed=0) as net:
        net.config[nengo.Ensemble].neuron_type = neuron_type
        net.config[nengo.Ensemble].gain = nengo.dists.Choice([1, -1])
        net.config[nengo.Ensemble].bias = nengo.dists.Uniform(-1, 1)

        net.integ = nengo.networks.EnsembleArray(neurons_per_d, dimensions)
        nengo.Connection(net.integ.output, net.integ.input, synapse=0.01)

        net.inp = nengo.Node([0] * dimensions)
        nengo.Connection(net.inp, net.integ.input, transform=0.01)

        net.p = nengo.Probe(net.integ.output)

    return net


def pes(dimensions, neurons_per_d, neuron_type):
    """
    PES learning rule benchmark.

    Parameters
    ----------
    dimensions : int
        Number of dimensions for vector values
    neurons_per_d : int
        Number of neurons to use per vector dimension
    neuron_type : `~nengo.neurons.NeuronType`
        Simulation neuron type

    Returns
    -------
    net : `nengo.Network`
        benchmark network
    """

    with nengo.Network(label="pes", seed=0) as net:
        net.config[nengo.Ensemble].neuron_type = neuron_type
        net.config[nengo.Ensemble].gain = nengo.dists.Choice([1, -1])
        net.config[nengo.Ensemble].bias = nengo.dists.Uniform(-1, 1)

        net.inp = nengo.Node([1] * dimensions)
        net.pre = nengo.Ensemble(neurons_per_d * dimensions, dimensions)
        net.post = nengo.Node(size_in=dimensions)

        nengo.Connection(net.inp, net.pre)

        conn = nengo.Connection(net.pre, net.post, learning_rule_type=nengo.PES())

        nengo.Connection(net.post, conn.learning_rule, transform=-1)
        nengo.Connection(net.inp, conn.learning_rule)

        net.p = nengo.Probe(net.post)

    return net


def basal_ganglia(dimensions, neurons_per_d, neuron_type):
    """
    Basal ganglia network benchmark.

    Parameters
    ----------
    dimensions : int
        Number of dimensions for vector values
    neurons_per_d : int
        Number of neurons to use per vector dimension
    neuron_type : `~nengo.neurons.NeuronType`
        Simulation neuron type

    Returns
    -------
    net : `nengo.Network`
        benchmark network
    """

    with nengo.Network(label="basal_ganglia", seed=0) as net:
        net.config[nengo.Ensemble].neuron_type = neuron_type

        net.inp = nengo.Node([1] * dimensions)
        net.bg = nengo.networks.BasalGanglia(dimensions, neurons_per_d)
        nengo.Connection(net.inp, net.bg.input)
        net.p = nengo.Probe(net.bg.output)

    return net


def mnist(use_tensor_layer=True):
    """
    A network designed to stress-test tensor layers (based on mnist net).

    Parameters
    ----------
    use_tensor_layer : bool
        If True, use individual tensor_layers to build the network, as opposed
        to a single TensorNode containing all layers.

    Returns
    -------
    net : `nengo.Network`
        benchmark network
    """

    with nengo.Network() as net:
        # create node to feed in images
        net.inp = nengo.Node(np.ones(28 * 28))

        if use_tensor_layer:
            nengo_nl = nengo.RectifiedLinear()

            ensemble_params = dict(
                max_rates=nengo.dists.Choice([100]), intercepts=nengo.dists.Choice([0])
            )
            amplitude = 1
            synapse = None

            x = nengo_dl.tensor_layer(
                net.inp,
                tf_compat.layers.conv2d,
                shape_in=(28, 28, 1),
                filters=32,
                kernel_size=3,
            )
            x = nengo_dl.tensor_layer(x, nengo_nl, **ensemble_params)

            x = nengo_dl.tensor_layer(
                x,
                tf_compat.layers.conv2d,
                shape_in=(26, 26, 32),
                transform=amplitude,
                filters=32,
                kernel_size=3,
            )
            x = nengo_dl.tensor_layer(x, nengo_nl, **ensemble_params)

            x = nengo_dl.tensor_layer(
                x,
                tf_compat.layers.average_pooling2d,
                shape_in=(24, 24, 32),
                synapse=synapse,
                transform=amplitude,
                pool_size=2,
                strides=2,
            )

            x = nengo_dl.tensor_layer(x, tf_compat.layers.dense, units=128)
            x = nengo_dl.tensor_layer(x, nengo_nl, **ensemble_params)

            x = nengo_dl.tensor_layer(
                x, tf_compat.layers.dropout, rate=0.4, transform=amplitude
            )

            x = nengo_dl.tensor_layer(x, tf_compat.layers.dense, units=10)
        else:
            nl = tf.nn.relu

            # def softlif_layer(x, sigma=1, tau_ref=0.002, tau_rc=0.02,
            #                   amplitude=1):
            #     # x -= 1
            #     z = tf.nn.softplus(x / sigma) * sigma
            #     z += 1e-10
            #     rates = amplitude / (tau_ref + tau_rc * tf.log1p(1 / z))
            #     return rates

            @nengo_dl.reshaped((28, 28, 1))
            def mnist_node(_, x):  # pragma: no cover
                x = tf_compat.layers.conv2d(x, filters=32, kernel_size=3, activation=nl)
                x = tf_compat.layers.conv2d(x, filters=32, kernel_size=3, activation=nl)
                x = tf_compat.layers.average_pooling2d(x, pool_size=2, strides=2)
                x = tf_compat.layers.flatten(x)
                x = tf_compat.layers.dense(x, 128, activation=nl)
                x = tf_compat.layers.dropout(x, rate=0.4)
                x = tf_compat.layers.dense(x, 10)

                return x

            node = nengo_dl.TensorNode(mnist_node, size_in=28 * 28, size_out=10)
            x = node
            nengo.Connection(net.inp, node, synapse=None)

        net.p = nengo.Probe(x)

    return net


def spaun(dimensions):
    """
    Builds the Spaun network from [1]_

    Parameters
    ----------
    dimensions : int
        Number of dimensions for vector values

    Returns
    -------
    net : `nengo.Network`
        benchmark network

    References
    ----------
    .. [1] Chris Eliasmith, Terrence C. Stewart, Xuan Choo, Trevor Bekolay,
       Travis DeWolf, Yichuan Tang, and Daniel Rasmussen (2012). A large-scale
       model of the functioning brain. Science, 338:1202-1205.

    Notes
    -----
    This network needs to be installed via

    .. code-block:: bash

        pip install git+https://github.com/drasmuss/spaun2.0.git
    """

    # pylint: disable=import-outside-toplevel
    from _spaun.configurator import cfg
    from _spaun.vocabulator import vocab
    from _spaun.experimenter import experiment
    from _spaun.modules.stim import stim_data
    from _spaun.modules.vision import vis_data
    from _spaun.modules.motor import mtr_data
    from _spaun.spaun_main import Spaun

    vocab.sp_dim = dimensions
    cfg.mtr_arm_type = None

    cfg.set_seed(1)
    experiment.initialize(
        "A",
        stim_data.get_image_ind,
        stim_data.get_image_label,
        cfg.mtr_est_digit_response_time,
        "",
        cfg.rng,
    )
    vocab.initialize(stim_data.stim_SP_labels, experiment.num_learn_actions, cfg.rng)
    vocab.initialize_mtr_vocab(mtr_data.dimensions, mtr_data.sps)
    vocab.initialize_vis_vocab(vis_data.dimensions, vis_data.sps)

    return Spaun()


def random_network(
    dimensions,
    neurons_per_d,
    neuron_type,
    n_ensembles,
    connections_per_ensemble,
    seed=0,
):
    """
    Basal ganglia network benchmark.

    Parameters
    ----------
    dimensions : int
        Number of dimensions for vector values
    neurons_per_d : int
        Number of neurons to use per vector dimension
    neuron_type : `~nengo.neurons.NeuronType`
        Simulation neuron type
    n_ensembles : int
        Number of ensembles in the network
    connections_per_ensemble : int
        Outgoing connections from each ensemble

    Returns
    -------
    net : `nengo.Network`
        benchmark network
    """

    random.seed(seed)
    with nengo.Network(label="random", seed=seed) as net:
        net.inp = nengo.Node([0] * dimensions)
        net.out = nengo.Node(size_in=dimensions)
        net.p = nengo.Probe(net.out)
        ensembles = [
            nengo.Ensemble(
                neurons_per_d * dimensions, dimensions, neuron_type=neuron_type
            )
            for _ in range(n_ensembles)
        ]
        dec = np.ones((neurons_per_d * dimensions, dimensions))
        for ens in net.ensembles:
            # add a connection to input and output node, so we never have
            # any "orphan" ensembles
            nengo.Connection(net.inp, ens)
            nengo.Connection(ens, net.out, solver=nengo.solvers.NoSolver(dec))

            posts = random.sample(ensembles, connections_per_ensemble)
            for post in posts:
                nengo.Connection(ens, post, solver=nengo.solvers.NoSolver(dec))

    return net


def run_profile(net, train=False, n_steps=150, do_profile=True, reps=1, **kwargs):
    """
    Run profiler on a benchmark network.

    Parameters
    ----------
    net : `~nengo.Network`
        The nengo Network to be profiled.
    train : bool
        If True, profile the ``sim.train`` function. Otherwise, profile the
        ``sim.run`` function.
    n_steps : int
        The number of timesteps to run the simulation.
    do_profile : bool
        Whether or not to run profiling
    reps : int
        Repeat the run this many times (only profile data from the last
        run will be kept).

    Returns
    -------
    exec_time : float
        Time (in seconds) taken to run the benchmark, taking the minimum over
        ``reps``.

    Notes
    -----
    kwargs will be passed on to `.Simulator`
    """

    exec_time = 1e10

    with net:
        nengo_dl.configure_settings(inference_only=not train)

    with nengo_dl.Simulator(net, **kwargs) as sim:
        if train:
            opt = tf_compat.train.GradientDescentOptimizer(0.001)
            x = np.random.randn(sim.minibatch_size, n_steps, net.inp.size_out)
            y = np.random.randn(sim.minibatch_size, n_steps, net.p.size_in)

            # run once to eliminate startup overhead
            sim.train(
                {net.inp: x, net.p: y}, optimizer=opt, n_epochs=1, profile=do_profile
            )

            for _ in range(reps):
                start = time.time()
                sim.train(
                    {net.inp: x, net.p: y},
                    optimizer=opt,
                    n_epochs=1,
                    profile=do_profile,
                )
                exec_time = min(time.time() - start, exec_time)
            print("Execution time:", exec_time)

        else:
            # run once to eliminate startup overhead
            sim.run_steps(n_steps, profile=do_profile)

            for _ in range(reps):
                start = time.time()
                sim.run_steps(n_steps, profile=do_profile)
                exec_time = min(time.time() - start, exec_time)
            print("Execution time:", exec_time)

    return exec_time


@click.group(chain=True)
def main():
    """Command-line interface for benchmarks."""


@main.command()
@click.pass_obj
@click.option("--benchmark", default="cconv", help="Name of benchmark network")
@click.option("--dimensions", default=128, help="Number of dimensions")
@click.option("--neurons_per_d", default=64, help="Neurons per dimension")
@click.option("--neuron_type", default="RectifiedLinear", help="Nengo neuron model")
@click.option(
    "--kwarg",
    type=str,
    multiple=True,
    help="Arbitrary kwarg to pass to benchmark network (key=value)",
)
def build(obj, benchmark, dimensions, neurons_per_d, neuron_type, kwarg):
    """Builds one of the benchmark networks"""

    # get benchmark network by name
    benchmark = globals()[benchmark]

    # get the neuron type object from string class name
    try:
        neuron_type = getattr(nengo, neuron_type)()
    except AttributeError:
        neuron_type = getattr(nengo_dl, neuron_type)()

    # set up kwargs
    kwargs = dict((k, int(v)) for k, v in [a.split("=") for a in kwarg])

    # add the special cli kwargs if applicable; note we could just do
    # everything through --kwarg, but it is convenient to have a
    # direct option for the common arguments
    params = inspect.signature(benchmark).parameters
    for kw in ("benchmark", "dimensions", "neurons_per_d", "neuron_type"):
        if kw in params:
            kwargs[kw] = locals()[kw]

    # build benchmark and add to context for chaining
    print(
        "Building %s with %s"
        % (nengo_dl.utils.function_name(benchmark, sanitize=False), kwargs)
    )

    obj["net"] = benchmark(**kwargs)


@main.command()
@click.pass_obj
@click.option(
    "--train/--no-train",
    default=False,
    help="Whether to profile training (as opposed to running) the network",
)
@click.option(
    "--n_steps", default=150, help="Number of steps for which to run the simulation"
)
@click.option("--batch_size", default=1, help="Number of inputs to the model")
@click.option(
    "--device",
    default="/gpu:0",
    help="TensorFlow device on which to run the simulation",
)
@click.option(
    "--unroll", default=25, help="Number of steps for which to unroll the simulation"
)
@click.option(
    "--time-only",
    is_flag=True,
    default=False,
    help="Only count total time, rather than profiling internals",
)
def profile(obj, train, n_steps, batch_size, device, unroll, time_only):
    """Runs profiling on a network (call after 'build')"""

    if "net" not in obj:
        raise ValueError("Must call `build` before `profile`")

    obj["time"] = run_profile(
        obj["net"],
        do_profile=not time_only,
        train=train,
        n_steps=n_steps,
        minibatch_size=batch_size,
        device=device,
        unroll_simulation=unroll,
    )


@main.command()
def matmul_vs_reduce():  # pragma: no cover
    """
    Compares two different approaches to batched matrix multiplication
    (tf.matmul vs tf.multiply+tf.reduce_sum).

    This is relevant for figuring out which approach is more efficient
    on a given system for different matrix shapes (determining which method
    we use in DotIncBuilder).
    """

    # a_shape = (n_ops, s0, s1, 1)
    # x_shape = (n_ops, 1, s1, mini)
    # for matmul we omit the 1 dimensions

    a_c = tf_compat.placeholder(tf.float64, shape=(None, None, None), name="a_c")
    x_c = tf_compat.placeholder(tf.float64, shape=(None, None, None), name="b_c")
    a_d = tf_compat.placeholder(tf.float64, shape=(None, None, None, 1), name="a_d")
    x_d = tf_compat.placeholder(tf.float64, shape=(None, 1, None, None), name="b_d")
    c = tf.matmul(a_c, x_c)
    d = tf.reduce_sum(input_tensor=tf.multiply(a_d, x_d), axis=-2)

    reps = 100
    n_ops_range = [1, 4, 8, 16, 32, 64]
    mini_range = [1, 16, 32, 64, 128]
    s0_range = [1, 64, 128, 192, 256]
    s1_range = [1, 64, 128, 192, 256]

    matmul_times = np.zeros(
        (len(n_ops_range), len(mini_range), len(s0_range), len(s1_range))
    )
    reduce_times = np.zeros_like(matmul_times)

    params = itertools.product(
        enumerate(n_ops_range),
        enumerate(mini_range),
        enumerate(s0_range),
        enumerate(s1_range),
    )

    with tf_compat.Session() as sess:
        for (i, n_ops), (j, mini), (k, s0), (l, s1) in params:
            print(n_ops, mini, s0, s1)

            a_val = np.random.randn(n_ops, s0, s1, 1)
            x_val = np.random.randn(n_ops, 1, s1, mini)

            for r in range(reps + 3):
                if r == 3:
                    start = time.time()
                c_val = sess.run(c, feed_dict={a_c: a_val[..., 0], x_c: x_val[:, 0]})
            matmul_times[i, j, k, l] = (time.time() - start) / reps

            for r in range(reps + 3):
                if r == 3:
                    start = time.time()
                d_val = sess.run(d, feed_dict={a_d: a_val, x_d: x_val})
            reduce_times[i, j, k, l] = (time.time() - start) / reps

            assert np.allclose(c_val, d_val)

    fig, ax = plt.subplots(len(n_ops_range), len(mini_range), sharex=True, sharey=True)

    X, Y = np.meshgrid(s0_range, s1_range)
    Z = matmul_times - reduce_times
    v = np.sort(np.concatenate((np.linspace(np.min(Z), np.max(Z), 10), [0])))
    for i, n_ops in enumerate(n_ops_range):
        for j, mini in enumerate(mini_range):
            cs = ax[i][j].contourf(X, Y, Z[i, j], v)
            if i == 0:
                ax[i][j].set_title("mini %d" % mini)
            if j == 0:
                ax[i][j].set_ylabel("ops %d" % n_ops)

    DATA_DIR = os.path.join(os.path.dirname(nengo_dl.__file__), "..", "data")
    np.savez(
        os.path.join(DATA_DIR, "matmul_benchmarks"),
        n_ops_range,
        mini_range,
        s0_range,
        s1_range,
        Z,
    )

    fig.subplots_adjust(right=0.8)
    cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
    fig.colorbar(cs, cax=cbar_ax)

    plt.show()


if __name__ == "__main__":
    main(obj={})  # pragma: no cover
